@@ -18,12 +18,14 @@ mod imp {
     use glib::{clone, ControlFlow, Properties};
     use gst::bus::BusWatchGuard;
     use gst::prelude::*;
+    use gtk::gdk::{Key, ModifierType};
     use gtk::prelude::*;
     use gtk::{gdk, CompositeTemplate};
     use tracing::{instrument, Span};
 
     use super::*;
     use crate::application::VaDisplayState;
+    use crate::glycin;
     use crate::texture_paintable::TexturePaintable;
     use crate::utils::gettext_f;
 
@@ -41,6 +43,12 @@ mod imp {
         scrolled_window: TemplateChild<gtk::ScrolledWindow>,
         #[template_child]
         title_label: TemplateChild<gtk::Label>,
+        #[template_child]
+        context_menu: TemplateChild<gtk::PopoverMenu>,
+        #[template_child]
+        click_menu_gesture: TemplateChild<gtk::GestureClick>,
+        #[template_child]
+        press_menu_gesture: TemplateChild<gtk::GestureLongPress>,
 
         #[property(get, set, construct_only)]
         application: OnceCell<Application>,
@@ -83,6 +91,8 @@ mod imp {
         resolution: PhantomData<String>,
         #[property(get, set)]
         show_overlay: Cell<bool>,
+        #[property(set = Self::set_menu_model)]
+        menu_model: PhantomData<Option<gio::MenuModel>>,
 
         constructed_at: OnceCell<Instant>,
         bus_watch_guard: RefCell<Option<BusWatchGuard>>,
@@ -110,6 +120,11 @@ mod imp {
                 klass.add_binding_signal(key, gdk::ModifierType::empty(), "activate");
             }
 
+            klass.install_action("menu.popup", None, |obj, _, _| {
+                obj.imp().show_context_menu(None);
+            });
+            klass.add_binding_action(Key::F10, ModifierType::SHIFT_MASK, "menu.popup");
+
             klass.bind_template();
             klass.bind_template_callbacks();
             klass.bind_template_instance_callbacks();
@@ -141,6 +156,10 @@ mod imp {
                     Signal::builder("stop-kinetic-scrolling")
                         .param_types([super::Picture::static_type()])
                         .build(),
+                    Signal::builder("setup-menu")
+                        .param_types([bool::static_type()])
+                        .action()
+                        .build(),
                 ]
             })
         }
@@ -190,6 +209,45 @@ mod imp {
             ));
             obj.add_controller(gesture);
 
+            // Click to open the context menu.
+            self.click_menu_gesture.connect_pressed(clone!(
+                #[weak(rename_to = imp)]
+                self,
+                move |gesture, _n_clicks, x, y| {
+                    if gesture.current_event().unwrap().triggers_context_menu() {
+                        gesture.set_state(gtk::EventSequenceState::Claimed);
+                        imp.show_context_menu(Some((x as i32, y as i32)));
+                    }
+                }
+            ));
+
+            // Touch long press to open the context menu.
+            self.press_menu_gesture.connect_pressed(clone!(
+                #[weak(rename_to = imp)]
+                self,
+                move |gesture, x, y| {
+                    gesture.set_state(gtk::EventSequenceState::Claimed);
+                    imp.show_context_menu(Some((x as i32, y as i32)));
+                }
+            ));
+
+            // Notification that the context menu closed.
+            self.context_menu.connect_notify_local(
+                Some("visible"),
+                clone!(
+                    #[weak]
+                    obj,
+                    move |menu, _| {
+                        if !menu.is_visible() {
+                            // Run in an idle because otherwise it happens before the actual action.
+                            glib::idle_add_local_once(move || {
+                                obj.emit_by_name::<()>("setup-menu", &[&false]);
+                            });
+                        }
+                    }
+                ),
+            );
+
             glib::MainContext::default().spawn_local(clone!(
                 #[weak(rename_to = imp)]
                 self,
@@ -330,6 +388,10 @@ mod imp {
                 .unwrap_or_else(|| gettext("N/A"))
         }
 
+        fn set_menu_model(&self, val: Option<&gio::MenuModel>) {
+            self.context_menu.set_menu_model(val);
+        }
+
         pub fn reset_kinetic_scrolling(&self, except_picture: Option<&Picture>) {
             if except_picture == Some(&*self.picture) {
                 return;
@@ -343,6 +405,20 @@ mod imp {
             self.scrolled_window.queue_allocate();
         }
 
+        fn show_context_menu(&self, xy: Option<(i32, i32)>) {
+            self.obj().emit_by_name::<()>("setup-menu", &[&true]);
+
+            let context_menu = &self.context_menu;
+            if let Some((x, y)) = xy {
+                let rect = gdk::Rectangle::new(x, y, 0, 0);
+                context_menu.set_pointing_to(Some(&rect));
+            } else {
+                context_menu.set_pointing_to(None);
+            }
+
+            context_menu.popup();
+        }
+
         pub fn grab_focus_(&self) {
             self.scrolled_window.grab_focus();
         }
@@ -388,31 +464,31 @@ mod imp {
             // much longer. However, due to GStreamer being prone to crashing (e.g. simply loading
             // a 16-bit PNG crashes the whole application at the time of this writing), we cannot
             // load the playbin in parallel. Oh well...
-            let image = match glycin::Loader::new(file.clone()).load().await {
+            let loader = glycin::loader_new(&file);
+            let image = match glycin::loader_load_future(&loader).await {
                 Ok(image) => image,
-                Err(err) => {
-                    if err.is_out_of_memory() {
-                        warn!("glycin reported out of memory; aborting loading");
-                        self.set_error();
-                        return;
-                    }
+                Err(_err) => {
+                    // FIXME: Missing from libglycin.
+                    //
+                    // if err.is_out_of_memory() {
+                    //     warn!("glycin reported out of memory; aborting loading");
+                    //     self.set_error();
+                    //     return;
+                    // }
 
                     let elapsed = self
                         .constructed_at
                         .get()
                         .expect("unexpected unset `constructed_at`")
                         .elapsed();
-                    debug!(
-                        "glycin failed to load format {:?} after {elapsed:?}, trying GStreamer",
-                        err.unsupported_format(),
-                    );
+                    debug!("glycin failed to load image after {elapsed:?}, trying GStreamer");
 
                     self.prepare_playbin(file);
                     return;
                 }
             };
 
-            let frame = match image.next_frame().await {
+            let frame = match glycin::image_next_frame_future(&image).await {
                 Ok(frame) => frame,
                 Err(err) => {
                     warn!("error loading frame with glycin: {err}");
@@ -432,7 +508,7 @@ mod imp {
             let obj = self.obj();
             let _guard = obj.freeze_notify();
 
-            let texture = frame.texture();
+            let texture = glycin::frame_get_texture(&frame);
             let paintable = TexturePaintable::new(&texture);
             self.picture.set_paintable(Some(paintable));
             assert_eq!(self.texture.replace(Some(texture)), None);
@@ -442,10 +518,12 @@ mod imp {
 
             obj.notify_resolution();
 
-            if let Some(format_name) = image.format_name() {
-                self.container_format.replace(Some(format_name));
-                self.obj().notify_container_format();
-            }
+            // FIXME: Missing from libglycin.
+            //
+            // if let Some(format_name) = image.details().info_format_name() {
+            //     self.container_format.replace(Some(format_name.to_owned()));
+            //     self.obj().notify_container_format();
+            // }
 
             self.stack.set_visible_child_name("content");
         }
@@ -822,7 +900,9 @@ mod imp {
 }
 
 glib::wrapper! {
-    pub struct Page(ObjectSubclass<imp::Page>) @extends adw::Bin, gtk::Widget;
+    pub struct Page(ObjectSubclass<imp::Page>)
+        @extends adw::Bin, gtk::Widget,
+        @implements gtk::Accessible, gtk::Buildable, gtk::ConstraintTarget;
 }
 
 #[gtk::template_callbacks]
