@@ -89,13 +89,11 @@ mod imp {
     use std::fmt;
     use std::fs::File;
     use std::marker::PhantomData;
-    use std::os::fd::AsFd as _;
     use std::str::FromStr;
     use std::time::Duration;
 
     use adw::prelude::*;
     use adw::subclass::prelude::*;
-    use ashpd::desktop::open_uri::OpenDirectoryRequest;
     use glib::{clone, closure, ControlFlow, Properties, SignalHandlerId, SourceId};
     use gtk::gdk::{self, Key, ModifierType};
     use gtk::{glib, CompositeTemplate};
@@ -184,6 +182,7 @@ mod imp {
         page_bindings: RefCell<HashMap<Page, Vec<glib::Binding>>>,
         page_is_loading_notify_id: RefCell<HashMap<Page, SignalHandlerId>>,
         page_stop_kinetic_scrolling_id: RefCell<HashMap<Page, SignalHandlerId>>,
+        page_setup_menu_id: RefCell<HashMap<Page, SignalHandlerId>>,
         switch_to_content_source_id: RefCell<Option<SourceId>>,
 
         surface_signals: OnceCell<glib::SignalGroup>,
@@ -210,8 +209,8 @@ mod imp {
         #[property(get, set = Self::set_selected_page, explicit_notify)]
         selected_page: RefCell<Option<Page>>,
 
-        /// If a tab menu is open for a page, this is that page, otherwise `None`.
-        menu_page: RefCell<glib::WeakRef<adw::TabPage>>,
+        /// If a tab/context menu is open for a page, this is that page, otherwise `None`.
+        menu_page: RefCell<glib::WeakRef<Page>>,
 
         /// In tabbed display mode, this is the tab bar widget.
         tab_bar: RefCell<Option<adw::TabBar>>,
@@ -226,6 +225,9 @@ mod imp {
         //
         // https://gitlab.gnome.org/GNOME/gtk/-/issues/7030
         pub open_file_dialog: RefCell<Option<glib::JoinHandle<()>>>,
+
+        /// Window identifier for portals (either Wayland handle or X11 window ID)
+        window_identifier: RefCell<Option<String>>,
     }
 
     #[glib::object_subclass]
@@ -345,7 +347,7 @@ mod imp {
             klass.install_action("win.about", None, |window, _, _| {
                 let resource_path = "/org/gnome/gitlab/YaLTeR/Identity/\
                                      org.gnome.gitlab.YaLTeR.Identity.metainfo.xml";
-                let about_window = adw::AboutDialog::from_appdata(resource_path, Some("25.03"));
+                let about_window = adw::AboutDialog::from_appdata(resource_path, Some("25.10"));
                 about_window.set_version(config::VERSION);
                 // Translators: shown in the About dialog, put your name here.
                 about_window.set_translator_credits(&gettext("translator-credits"));
@@ -366,9 +368,9 @@ mod imp {
                 // DL doesn't extract release notes from metainfo, so let's help it out with the
                 // ones shown in the dialog.
                 let gettext = |_| ();
-                gettext("This release has a minor visual refresh for GNOME 48.");
-                gettext("Changed the versioning scheme to year.month.");
-                gettext("Updated to the GNOME 48 platform.");
+                gettext("This release adds a context menu to row and column display modes and updates the platform to GNOME 49.");
+                gettext("Added a context menu on the images, mirroring the tab context menu. This makes the context menu accessible in row and column display modes where there are no tabs.");
+                gettext("Updated to the GNOME 49 platform.");
                 gettext("Updated translations.");
             });
         }
@@ -414,6 +416,38 @@ mod imp {
             ));
             self.surface_signals.set(surface_signals).unwrap();
 
+            obj.connect_map(move |obj| {
+                let Some(surface) = obj.native().and_then(|x| x.surface()) else {
+                    return;
+                };
+
+                if let Some(toplevel) = surface.downcast_ref::<gdk4_wayland::WaylandToplevel>() {
+                    toplevel.export_handle(clone!(
+                        #[weak]
+                        obj,
+                        move |_, res| {
+                            match res {
+                                Ok(handle) => {
+                                    obj.imp()
+                                        .window_identifier
+                                        .replace(Some(format!("wayland:{handle}")));
+                                }
+                                Err(err) => {
+                                    warn!("failed exporting Wayland handle: {err:?}")
+                                }
+                            }
+                        }
+                    ));
+                } else if let Some(x11_surface) = surface.downcast_ref::<gdk4_x11::X11Surface>() {
+                    obj.imp()
+                        .window_identifier
+                        .replace(Some(format!("x11:{:x}", x11_surface.xid())));
+                }
+            });
+            obj.connect_unmap(move |obj| {
+                obj.imp().window_identifier.replace(None);
+            });
+
             self.content_toolbar_view
                 .connect_reveal_bottom_bars_notify(clone!(
                     #[weak]
@@ -839,6 +873,10 @@ mod imp {
                     .bidirectional()
                     .sync_create()
                     .build(),
+                self.tab_view
+                    .bind_property("menu-model", &page, "menu-model")
+                    .sync_create()
+                    .build(),
             ];
             if self
                 .page_bindings
@@ -873,6 +911,35 @@ mod imp {
                 error!("`page_stop_kinetic_scrolling_id` already had an entry for this page");
             };
 
+            let id = page.connect_local(
+                "setup-menu",
+                false,
+                clone!(
+                    #[weak(rename_to = imp)]
+                    self,
+                    #[upgrade_or]
+                    None,
+                    move |args| {
+                        let page = args[0].get().unwrap();
+                        let is_open: bool = args[1].get().unwrap();
+                        if is_open {
+                            imp.setup_menu(Some(&page));
+                        } else {
+                            imp.setup_menu(None);
+                        }
+                        None
+                    }
+                ),
+            );
+            if self
+                .page_setup_menu_id
+                .borrow_mut()
+                .insert(page.clone(), id)
+                .is_some()
+            {
+                error!("`page_setup_menu_id` already had an entry for this page");
+            };
+
             if page.is_error() {
                 self.stack.set_visible_child_name("content");
                 self.obj().present_if_not_visible();
@@ -942,6 +1009,12 @@ mod imp {
                 error!("detached page should have `page_stop_kinetic_scrolling_id` entry");
             }
 
+            if let Some(id) = self.page_setup_menu_id.borrow_mut().remove(&page) {
+                page.disconnect(id);
+            } else {
+                error!("detached page should have `page_setup_menu_id` entry");
+            }
+
             page.reset_kinetic_scrolling(None);
 
             if let Some(playbin) = page.playbin() {
@@ -1007,13 +1080,6 @@ mod imp {
 
         #[template_callback]
         fn on_setup_menu(&self, tab_page: Option<&adw::TabPage>) {
-            debug!("setup-menu {tab_page:?}");
-
-            let page = tab_page
-                .map(|tab_page| tab_page.downgrade())
-                .unwrap_or_default();
-            self.menu_page.replace(page);
-
             if let Some(tab_page) = tab_page {
                 let page: Page = tab_page
                     .child()
@@ -1023,7 +1089,19 @@ mod imp {
                     .expect("tab page child bin has no child")
                     .downcast()
                     .expect("tab page child bin child has wrong type");
+                self.setup_menu(Some(&page));
+            } else {
+                self.setup_menu(None);
+            }
+        }
 
+        fn setup_menu(&self, page: Option<&Page>) {
+            debug!("setup_menu {page:?}");
+
+            let weak = page.map(|page| page.downgrade()).unwrap_or_default();
+            self.menu_page.replace(weak);
+
+            if let Some(page) = page {
                 let has_path = page.file().path().is_some();
                 self.obj().action_set_enabled("win.show-in-files", has_path);
             } else {
@@ -1032,26 +1110,33 @@ mod imp {
         }
 
         fn menu_or_selected_tab_page(&self) -> Option<adw::TabPage> {
-            self.menu_page
-                .borrow()
-                .upgrade()
-                .or_else(|| self.tab_view.selected_page())
-        }
-
-        fn menu_or_selected_page(&self) -> Option<Page> {
-            self.menu_page
-                .borrow()
-                .upgrade()
-                .map(|tab_page| {
-                    tab_page
+            if let Some(menu_page) = self.menu_page.borrow().upgrade() {
+                (0..self.tab_view.n_pages()).find_map(|i| {
+                    let tab_page = self.tab_view.nth_page(i);
+                    let page: Page = tab_page
                         .child()
                         .downcast::<adw::Bin>()
                         .expect("tab page child has wrong type")
                         .child()
                         .expect("tab page child bin has no child")
                         .downcast()
-                        .expect("tab page child bin child has wrong type")
+                        .expect("tab page child bin child has wrong type");
+
+                    if page == menu_page {
+                        Some(tab_page)
+                    } else {
+                        None
+                    }
                 })
+            } else {
+                self.tab_view.selected_page()
+            }
+        }
+
+        fn menu_or_selected_page(&self) -> Option<Page> {
+            self.menu_page
+                .borrow()
+                .upgrade()
                 .or_else(|| self.selected_page())
         }
 
@@ -1072,6 +1157,11 @@ mod imp {
                 return;
             };
 
+            let Some(window_identifier) = self.window_identifier.borrow().clone() else {
+                warn!("no window identifier available for portal call");
+                return;
+            };
+
             // The OpenDirectory portal wants a file descriptor. There's a way to get it without
             // leaving gio:
             //
@@ -1106,18 +1196,8 @@ mod imp {
                 }
             };
 
-            let Some(native) = self.obj().native() else {
-                warn!("self.obj().native() returned None");
-                return;
-            };
-
-            let identifier = ashpd::WindowIdentifier::from_native(&native).await;
-            if let Err(err) = OpenDirectoryRequest::default()
-                .identifier(identifier)
-                .send(&file.as_fd())
-                .await
-            {
-                warn!("OpenDirectory returned an error: {:?}", err);
+            if let Err(err) = call_open_directory_portal(window_identifier, file).await {
+                warn!("failed to call OpenDirectory: {err:?}");
             }
         }
 
@@ -1147,17 +1227,41 @@ mod imp {
         }
 
         pub fn move_tab_to_new_window(&self) {
-            if let Some(page) = self.menu_or_selected_tab_page() {
-                let application: Application = self
-                    .obj()
-                    .application()
-                    .expect("application was not set")
-                    .downcast()
-                    .expect("application has wrong type");
-                let new_window = application.create_new_window();
-                self.tab_view
-                    .transfer_page(&page, &new_window.imp().tab_view, 0);
-                new_window.present();
+            let application: Application = self
+                .obj()
+                .application()
+                .expect("application was not set")
+                .downcast()
+                .expect("application has wrong type");
+
+            match self.display_mode.get() {
+                DisplayMode::Tabbed => {
+                    if let Some(page) = self.menu_or_selected_tab_page() {
+                        let new_window = application.create_new_window();
+                        let new_imp = new_window.imp();
+                        self.tab_view.transfer_page(&page, &new_imp.tab_view, 0);
+                        new_window.present();
+                    }
+                }
+                DisplayMode::Row | DisplayMode::Column => {
+                    if let Some(page) = self.menu_or_selected_page() {
+                        let new_window = application.create_new_window();
+                        let new_imp = new_window.imp();
+                        new_imp.set_display_mode(self.display_mode.get());
+
+                        self.page_grid.close_page(&page);
+                        self.on_page_detached(page.clone());
+
+                        if self.page_grid.n_pages() == 0 {
+                            self.stack.set_visible_child_name("empty");
+                        }
+
+                        new_imp.page_grid.append(page.clone());
+                        new_imp.on_page_attached(page);
+
+                        new_window.present();
+                    }
+                }
             }
         }
 
@@ -1577,6 +1681,31 @@ mod imp {
         }
     }
 
+    async fn call_open_directory_portal(
+        window_identifier: String,
+        file: File,
+    ) -> Result<(), Box<dyn std::error::Error>> {
+        let bus = gio::bus_get_future(gio::BusType::Session).await?;
+
+        let fd_list = gio::UnixFDList::from_array([file]);
+        let options: HashMap<String, glib::Variant> = HashMap::new();
+
+        bus.call_with_unix_fd_list_future(
+            Some("org.freedesktop.portal.Desktop"),
+            "/org/freedesktop/portal/desktop",
+            "org.freedesktop.portal.OpenURI",
+            "OpenDirectory",
+            Some(&(window_identifier, glib::variant::Handle(0), options).to_variant()),
+            None,
+            gio::DBusCallFlags::NONE,
+            -1,
+            Some(&fd_list),
+        )
+        .await?;
+
+        Ok(())
+    }
+
     #[cfg(test)]
     mod tests {
         use super::*;
@@ -1609,7 +1738,8 @@ mod imp {
 glib::wrapper! {
     pub struct Window(ObjectSubclass<imp::Window>)
         @extends adw::ApplicationWindow, gtk::ApplicationWindow, gtk::Window, gtk::Widget,
-        @implements gio::ActionGroup, gio::ActionMap;
+        @implements gtk::Accessible, gtk::Buildable, gtk::ConstraintTarget,
+            gtk::Native, gtk::Root, gtk::ShortcutManager, gio::ActionGroup, gio::ActionMap;
 }
 
 #[gtk::template_callbacks]
